;; $Id: blocker.lisp,v 1.17 2005/05/23 21:11:27 jimka Exp $

#|

 This software is Copyright (c) 2005 Jim Newton <jimka@rdrop.com>

 Jim Newton grants you the rights to distribute
 and use this software as governed by the terms
 of the Lisp Lesser GNU Public License
 (http://opensource.franz.com/preamble.html),
 known as the LLGPL.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
|#

(in-package blocker)


#+clisp (setf *WARN-ON-FLOATING-POINT-CONTAGION* nil)

(defmacro exists ( var some-list &rest sexpr)
  `(member-if (lambda ( ,var) ,@sexpr) ,some-list))

(defmacro setof ( var some-list &rest sexprs)
  `(loop for ,var in ,some-list
    when (progn ,@sexprs)
    collect ,var))

(defmacro print-vals (&rest body)
  `(print-vals-internal ,(length body) 1 ,body))

(defmacro print-vals-internal (len depth body)
  (when body
    `(progn (format t "~D/~D: ~A = ~A~%"
	     ,depth ,len
	     ',(car body)
	     ,(car body))
      (print-vals-internal
        ,len
       ,(1+ depth)
       ,(cdr body)))))




(defmacro make-direction (dir)
  `(let (enclosed_dir)
    (defun ,dir ()
      (or enclosed_dir
	  (setf enclosed_dir (make-instance ',dir))))))

(make-direction northwest)
(make-direction southwest)
(make-direction northeast)
(make-direction southeast)
(make-direction east)
(make-direction west)
(make-direction south)
(make-direction north)
(make-direction no-direction)

(defparameter *state-caches* nil)

(defun clear-cache ()
  (dolist ( cache *state-caches*)
    (format t "clearing hash count=~A size=~A~%"
	    (hash-table-count cache)
	    (hash-table-size cache))
    (clrhash cache)))

(defmacro def-cached-method ( method-name lambda-list get-key &rest body)
  `(progn
    (let (( cache (make-hash-table :test #'equal)))

      (unless (find cache *state-caches*)
	(push cache *state-caches*))

      (defmethod ,method-name :around ,lambda-list
	(let* (( key ,get-key)
	       ( value (gethash key cache)))
	  (cond
	    ( value
	     (car value))
	    (t
	     (setf value (call-next-method)
		   (gethash key cache) (list value))

	     value))))

      (defmethod ,method-name ,lambda-list
	,@body))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; we define the classes here because sbcl issues
;; warning when we reference classes before they are
;; defined.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defclass blocker-base nil nil)

(defclass board ( blocker-base )
  (( num-rows
     :initform 11
     :reader num-rows)
   ( num-cols
     :initform 11
     :reader num-cols)
   ( white
     :accessor white)
   ( black
     :accessor black)
   ( players
     :accessor players)
   ( squares
     :accessor squares)
   ( event-handlers
     :accessor event-handlers
     :initform nil)
   ( output-files
     :initform (list *standard-output*)
     :accessor output-files)))

(defclass piece-move ( blocker-base )
  (( piece
     :initarg :piece
     :reader piece)
   ( combination
     :initarg :combination
     :accessor combination)))


(defclass square ( blocker-base)
  (( row
     :initarg :row
     :accessor row)
   ( col
     :initarg :col
     :accessor col)
   ( occupant
     :initform nil
     :accessor occupant)))

(defclass combination ( blocker-base )
  (( directions
     :initarg :directions
     :accessor directions)))

(defclass direction ( blocker-base )
  (( symbol
     :accessor -symbol)
   ( name
     :accessor name)
   ( angle
     :accessor angle)
   ( components
     :initform nil
     :accessor components)
   ( deltax
     :initform 0
     :accessor deltax)
   ( deltay
     :initform 0
     :accessor deltay)
   ( swept-directions
     :initform nil
     :accessor swept-directions)))

(defclass manhattan ( direction )
  nil)

(defclass west ( manhattan )
  (( symbol
     :initform "w")
   ( name
     :initform "West")
   ( deltax
     :initform -1)))



(defclass east ( manhattan )
  (( symbol
     :initform "e")
   ( name
     :initform "East")
   ( deltax
     :initform 1)))


(defclass south ( manhattan )
  (( symbol
     :initform "s")
   ( name
     :initform "South")
   ( deltay
     :initform -1)))

(defclass north ( manhattan )
  (( symbol
     :initform "n")
   ( name
     :initform "North")
   ( deltay
     :initform 1)))

(defclass diagonal ( direction)
  nil)


(defclass northeast ( diagonal )
  (( symbol :initform "ne")
   ( name   :initform "Northeast")
   ( components :initform (list (north) (east)))))

(defclass northwest ( diagonal)
  (( symbol :initform "nw")
   ( name   :initform "Northwest")
   ( components :initform (list (north) (west)))))


(defclass southeast ( diagonal)
  (( symbol :initform "se")
   ( name   :initform "Southeast")
   ( components :initform (list (south) (east)))))


(defclass southwest ( diagonal )
  (( symbol :initform "sw")
   ( name   :initform "Southwest")
   ( components :initform (list (south) (west)))))


(defclass no-direction ( direction )
  (( symbol :initform "i")
   ( name :initform "identity")))

(defclass game ( blocker-base )
  (( history
     :initform nil
     :accessor history)
   ( game-over
     :initform nil
     :accessor game-over)
   ( verbose
     :initform t
     :initarg :verbose
     :accessor verbose)
   ( turn
     :initform nil
     :accessor turn)
   ( directions
     :accessor directions)
   ( computer
     :accessor computer)
   ( history-file
     :accessor history-file
     :initform nil)
   ( hist-suffix
     :initform "his"
     :accessor hist-suffix)
   ( history-dir
     :initform "history"
     :accessor history-dir)
   ( remaining-moves
     :initform 999
     :accessor remaining-moves)
   ( event-queue
     :initform (make-instance 'queue)
     :reader event-queue)))

(defclass blocker ( game board)
  (( events
     :accessor events)
   ( directions
     :initform (list (north) (northwest) (northeast)
		     (south) (southwest) (southeast)
		     (east) (west) (no-direction))
     :reader directions)))

(defclass blocker-event ( blocker-base )
  (( blocker
     :reader blocker
     :initarg :blocker)
   ( name
     :initform "none-yet"
     :accessor name)
   ( number
     :accessor -number)
   ( shortcut
     :initform "none-yet"
     :accessor shortcut)))


(defclass movement ( blocker-base )
  (( direction
     :reader direction
     :initarg :direction)
   ( piece
     :reader piece
     :initarg :piece)))

(defclass piece ( blocker-base )
  (( player
     :accessor player)
   ( name
     :accessor name)
   ( number-of-rolls
     :accessor number-of-rolls)
   ( volume
     :accessor volume)
   ( blocker
     :initarg :blocker
     :accessor blocker)
   ( directions
     :accessor directions)
   ( combinations
     :accessor combinations)
   ( eff-comb-table
     :reader eff-comb-table
     :initform (make-hash-table :test #'equal))
   ( state
     :accessor state)
   ( squares
     :initform nil
     :accessor squares)))


;; a regular piece does not change its shape as it moved.
;; e.g., a piece that is 1x1x2 might become 1x2x1 when it rolls
;;   so it is irregular.
;; but a pice that is 2x2x2, remains 2x2x2 as it rolls.
;; also the godfather which is 1x1x0 does not roll, but slides
;; thus it remains the same shape, it is regular as well.
(defclass regular-piece ( piece)
  nil)

(defclass rollable ( piece )
  nil)

(defclass state ( blocker-base )
  (( piece
     :type piece
     :initarg :piece
     :accessor piece)
   ( x-coord
     :type (signed-byte 8)
     :initarg :x-coord
     :accessor x-coord)
   ( y-coord
     :type (signed-byte 8)
     :initarg :y-coord
     :accessor y-coord)
   ( dx
     :type (signed-byte 8)
     :initarg :dx
     :accessor dx)
   ( dy
     :type (signed-byte 8)
     :initarg :dy
     :accessor dy)
   ( dz
     :type (signed-byte 8)
     :initarg :dz
     :accessor dz)))

(defclass strategy ( blocker-base)
  (( player
     :initarg :player
     :accessor player)
   ( blocker
     :initarg :blocker)
   ( components
     :accessor components)
   ( suffix
     :initform "str"
     :accessor suffix)
   ( strategy-dir
     :initform "strategy"
     :accessor strategy-dir)
   ( strategy-base
     :initform "strat.clo"
     :accessor strategy-base)
   ( lock-name
     :accessor lock-name)))


(defclass player ( blocker-base )
  (( name
     :accessor name)
   ( symbol
     :accessor -symbol)
   ( blocker
     :initarg :blocker
     :accessor blocker)
   ( input-file
     :initform *standard-input*
     :accessor input-file)
   ( pieces
     :accessor pieces)
   ( rollable-pieces
     :accessor rollable-pieces)
   ( godfather
     :accessor godfather)
   ( thin
     :accessor thin)
   ( fat
     :accessor fat)
   ( flat
     :accessor flat)
   ( frecher
     :accessor frecher)
   ( opponent
     :accessor opponent)
   ( home-row
     :accessor home-row)
   ( strategy
     :accessor strategy)

   ( selected-piece
     :initform nil
     :accessor selected-piece)
   ( selected-combination-conc
     :initform (list nil)
     :accessor selected-combination-conc)
   ( num-combinations
     :accessor num-combinations)))
(defclass queue ( blocker-base )
  (( conc
     :reader conc
     :initform (list nil))))

(defclass strategy-component ( blocker-base )
  (( strategy
     :initarg :strategy
     :accessor strategy)
   ( name
     :accessor name)
   ( coef
     :accessor coef)))


(defclass input-device ( blocker-base )
  nil)

(defclass ascii ( input-device )
  nil)





;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod initialize-instance :after (( self board) &rest ignored)
  (declare (ignore ignored))
  (setf (squares self)
	(make-array (list (num-rows self) (num-cols self))
		    :element-type 'square
		    :adjustable nil
		    :initial-contents
		    (loop for col from 0 below (num-cols self)
			  collect
			  (loop for row from 0 below (num-rows self)
				collect
				(make-instance 'square :col col :row row))))
	

	(black self)
	(make-instance 'black :blocker self))
  
  (initial-placement (black self))
  
  (setf (white self) (make-instance 'white :blocker self))
  (initial-placement (white self))

  (setf	(players self)            (list (white self) (black self))
	(opponent (white self)) (black self)
	(opponent (black self)) (white self))
  )


(defmethod initialize-instance :after (( self direction) &rest ignored)
  (declare (ignore ignored))
  (setf (angle self)
	(if (and (zerop (deltay self)) (zerop (deltax self)))
	    ;; CMUCL and SBCL already return (atan 0 0)-->0
	    ;; but CLISP signals an error.  so we simply return
	    ;; zero in this case because we really do not care about
	    ;; the angle anyway if rise=run=0	    
	    0
	    
	    (atan (deltay self) (deltax self)))))


(defmethod initialize-instance :after (( self game) &rest ignored)
  (declare (ignore ignored))
  ;;(format t "initializing game ~A~%" self )
  ;; collect all the events, each event is an instance of
  ;; some subclass of EVENT.  each has a different shortcut.
  (setf (events self) (loop for ev in '(
					no-event
					game-over
					quit
					valid-events
					valid-combinations
					move-0
					move-1
					move-2
					move-4
					move-8
					move-s
					move-n
					move-e
					move-w
					move-ne
					move-nw
					move-se
					move-sw
					move-i
					draw-board
					verbose-yes
					verbose-no
					undo
					suggest
					human-becomes-white
					human-becomes-black
					tk-becomes-white
					tk-becomes-black
					computer-becomes-white
					computer-becomes-black
					other-turn
					move-automatic
					auto-play
					read-strategy
					write-strategy
					
					)
			    collect (make-instance ev :blocker self))

	(turn self) (white self)
	(computer self) (list (black self))
  
	(history-file self) 

	(loop with index = 0
	      and  filename = "/dev/null"
	      if (probe-file filename)
	      do
	      (setf filename (format nil "~A/~A.~A" (history-dir self) index (hist-suffix self)))
	      (incf index)
		    
	      else return filename)))


(defmethod initialize-instance :after (( self blocker) &rest ignored)
  (declare (ignore ignored))
  (assert (null (member nil (directions self))))
  nil)


(defmethod initialize-instance :after (( self blocker-event ) &rest ignored)
  (declare (ignore ignored))
  ;;(format t "initializing event ~A shortcut=~A ~%"
  ;; (class-name (class-of self)) (shortcut self))
  nil)






(defmethod initialize-instance :after (( self piece) &rest ignored)
  (declare (ignore ignored))
  ;;(format t "initializing piece~%")
  (setf (state self) (make-instance 'state :piece self)))

(defmethod initialize-instance :after (( self rollable) &rest ignored)
  (declare (ignore ignored))

  (setf (directions self)
	(loop for x in (directions (blocker self))
	      unless (diagonal-p x)
	      collect x))
  (calc-combinations self))


(defmethod printable (( self square))
  (format nil "<r=~A,c=~A>" (row self) (col self)))


(defmethod initialize-instance :after (( self strategy) &rest ignored)
  (declare (ignore ignored))
  (setf (lock-name self ) (format nil "~A/lock" (strategy-dir self))
	(components self) (loop for sc in 
				'(
				  player-shadow
				  opp-shadow
				  wasted-shadow
				  check
				  encroachment
				  danger-of-check
				  back-line
				  gf-free-squares
				  escape-route
				  frecher-gf-dist
				  thin-gf-dist
				  flat-gf-dist
				  fat-gf-dist
				  rand)
				collect (make-instance sc :strategy self))))


(defmethod initialize-instance :after (( self player) &rest ignored)
  (declare (ignore ignored))
  ;;(format t "initializing player: ~A of ~A~%" self (blocker self))
  (setf (godfather self) (make-instance 'godfather :blocker (blocker self))
	(thin self)      (make-instance 'thin      :blocker (blocker self))
	(flat self)      (make-instance 'flat      :blocker (blocker self))
	(fat self)       (make-instance 'fat       :blocker (blocker self))
	(frecher self)   (make-instance 'frecher   :blocker (blocker self))
	(strategy self)  (make-instance 'strategy  :player    self)
	(rollable-pieces self) (list (frecher self)
				     (thin self)
				     (flat self)
				     (fat self))
	(pieces self) (cons (godfather self)
			    (rollable-pieces self))
	(num-combinations self) (apply #'+
				       (mapcar #'length (mapcar #'combinations (pieces self)))))
  (loop for piece in (pieces self)
	do (setf (player piece) self)))

(defmethod select-event (( self blocker))
  (let ( matching selection )
    (loop while (empty-p (event-queue self))
	  do
	  (write-output self (format nil "~A Event Selection: "
			       (printable (turn self))))
	  (setf selection (read-input self))
	  (when (setf matching (find-event-by-shortcut self selection))
	    (qpush (event-queue self) matching)
	    (write-output self (format nil "~A: you chose ~A: ~A~%"
				 (name (turn self))
				 (shortcut matching)
				 (prompt matching))))
	  
	  (when (empty-p (event-queue self))
	    (write-output self (format nil "Invalid event selected: ~A~%" selection))))))
	    

(defmethod elements (( q queue))
  (car (conc q)))

(defun tconc ( conc item)
  (let (( new (list item)))
    (cond
     ((car conc)
      (setf (cdr (cdr conc)) new))
     (t
      (setf (car conc) new)))

    (setf (cdr conc) new)))

(defmethod qpush (( q queue) item)
  (tconc (conc q) item))

(defmethod qpop (( q queue))
  (prog1 (car (elements q))
    (rplaca (conc q) (cdr (car (conc q))))
    (unless (elements q)
      (rplacd (conc q) nil))))
	
(defmethod empty-p (( q queue))
  (null (elements q)))

(defmethod qflush (( q queue))
  (loop until (empty-p q)
	do (qpop q)))
