;; $Id: game.lisp,v 1.12 2005/05/23 21:09:42 jimka Exp $

#|

 This software is Copyright (c) 2005 Jim Newton <jimka@rdrop.com>

 Jim Newton grants you the rights to distribute
 and use this software as governed by the terms
 of the Lisp Lesser GNU Public License
 (http://opensource.franz.com/preamble.html),
 known as the LLGPL.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
|#

(in-package blocker)


;; (xrange 3) --> ( 2 1 0)
(defmethod xrange (( self integer))
  (loop for x from 0 below self
	collect x))


(defmethod event-setup (event-handler)
  (format t "warning no event-setup specializing on ~A~%" event-handler))

(defmethod event-setup ((eh (eql *standard-input*)))
  (declare (ignore eh))
  nil)

(defmethod read-input (( self board))
  (event-setup (input-file (turn self)))
  (event-setup (input-file (opponent (turn self))))
  (read-input (input-file (turn self))))


(defmethod read-input (( self (eql *standard-input*)))
  (read-line))

(defmethod write-output (( self board) str)
  (event-setup (input-file (turn self)))
  (event-setup (input-file (opponent (turn self))))
  (loop for fp in (output-files self)
	do
	(princ str fp)
	(unless (equal fp *standard-output*)
	  (terpri))))

(defmethod size (( self board))
  (* (num-rows self)
     (num-cols self)))

(defmethod get-state-key (( self board))
  (format nil "~A" (loop for player in (players self)
			 nconcing
			(loop for piece in (pieces player)
			      collect (shorthand (state piece))))))

(defmethod new-output (( self board) fp)
  (unless (find fp (output-files self))
    (setf (output-files self) (cons fp (output-files self)))))

(defmethod forget-output (( self board) fp)
  (setf (output-files self) (remove fp (output-files self))))

(defmethod print-shadow (( self board))
  (print-shadow (black self))
  (print-shadow (white self)))

(defun is-valid-square ( board row col)
  (declare (type (signed-byte 8) row col)
	   (type board board))
  (and (<= 0 col)
       (< col (num-cols board))
       (<= 0 row)
       (< row (num-rows board))))

(defun get-square ( board row col)
  (declare (type (signed-byte 8) row col)
	   (type board board))
  (when (is-valid-square board row col)
    (the square (aref (squares board) col row))))

(defmethod get-occupant (( self (eql nil)) &key row col)
  (declare (ignore row col))
  nil)

(defmethod occupant (( self (eql nil)))
  nil)

(defmethod get-occupant (( self board) &key row col)
  ;; note that if get-square returns nil, then there
  ;; is a special get-occupant method specializing on nil
  (occupant (get-square self row col)))

(defmethod occupy (( self board) ( row integer) ( col integer) ( piece piece))
  (let (( square (get-square self row col)))
    (cond
      ((null square)
       (format t "ERROR no such square row=~A col=~A~%"
	       row col)
       (assert nil))

      ((occupant square)
       (new-output self *standard-output*)
       (draw self)
       (format t "ERROR ~A ~A ~A~%"
	       square piece (occupant square))
       (assert (null (occupant square))))

      (t
       (setf (occupant square) piece)
       (setf (squares piece) (cons square (squares piece)))))))

(defmethod update (( self board ) ( piece piece))
  (let (( state (state piece)))
    (assert (null (squares piece)))
    (dotimes ( col (dx state))
      (dotimes ( row (dy state))
	(occupy self
		(+ (y-coord state) row)
		(+ (x-coord state) col)
		piece)))))

(defun eq* ( x &rest args)
  (or (null args)
      (and (eq x (car args))
	   (apply #'eq* args))))

(defmethod divide-rows (( self board) row1 row2)
  (write-output self "       ")
  (dotimes ( col (num-cols self))
    (write-output self
	    (if (and      (get-occupant self :row row1 :col col)
		     (eq* (get-occupant self :row row1 :col col)
			  (get-occupant self :row row2 :col col)
			  (get-occupant self :row row1 :col (1- col))
			  (get-occupant self :row row2 :col (1- col))))
		" "
		"+"))
    (write-output self
	    (if (and      (get-occupant self :row row1 :col col)
		     (eq* (get-occupant self :row row1 :col col)
			  (get-occupant self :row row2 :col col)))
		"  "
		"--")))
  (write-output self "+")
  (write-output self #\Newline))

(defmethod draw-row (( self board) y)
  (let ( occupant )
    (dotimes ( col (num-cols self))
      (write-output self
	      (if (and      (get-occupant self :row y :col col)
		       (eql (get-occupant self :row y :col col)
			    (get-occupant self :row y :col (1- col))))
		  " "
		  "|"))
      (setf occupant (get-occupant self :row y :col col))
      
      (cond
	((and occupant (= 2 (dz (state occupant))))
	 (write-output self (nth 1 (-symbol (player occupant))))
	 (write-output self (format nil "~A" (volume occupant))))
	
	( occupant
	 (write-output self (nth 0 (-symbol (player occupant))))
	 (write-output self (format nil "~A" (volume occupant))))
	
	(t
	 (write-output self "  "))))
    
    (write-output self "|")
    (write-output self #\Newline)))

(defmethod draw-col-nums (( self board))
  (write-output self "col=   ")
  (loop for col from 0 below (num-cols self)
	do
	(write-output self (format nil " ~2D" col)))
  (write-output self #\Newline))

(defmethod draw (( self board))
  ;; (write-output self (format nil "board state: ~A~%" (get-state-key self)))
  (draw-col-nums self)
  (loop for y
	from (1- (num-rows self))
	downto 0
	do
	(divide-rows self y (1+ y))
	(write-output self (format nil "row=~2D " y))
	(draw-row self y))
  (divide-rows self 0 -1)
  (draw-col-nums self)
  (finish-output))

        
(defmethod find-player (( self board) symbol)
  (loop for player in (players self)
	when (member symbol (-symbol player))
	do (return-from find-player player))
  nil)





(defmethod do-event (( self game) ( shortcut string))
  (put-event self shortcut)
  (do-pending-events self))

(defmethod do-event (( game game) ( event blocker-event))
  (put-event game event)
  (do-pending-events game))

(defmethod put-event (( game game) (shortcut string))
  (put-event game (find-event-by-shortcut game shortcut)))

(defmethod put-event (( game game) ( event blocker-event))
  (qpush (event-queue game) event))

(defmethod remember-move (( game game) ( direction direction) ( piece piece))
  (setf (history game) (cons (make-instance 'movement
					    :direction direction
					    :piece piece)
			     (history game))))

(defmethod forget-move (( game game))
  (pop (history game)))

(defmethod undo-move (( self game))
  (let (( movement (car (history self))))
    (unmove (direction movement) (piece movement))))

(defmethod undo-combination (( self game))
  (let (( player (player (piece (car (history self))))))
    (loop while (and (history self)
		     (eql player (player (piece (car (history self))))))
	  do
	  (format t "undo ~A~%" (printable (car (history self))))
	  (undo-move self))
    (other-turn self)
    (draw self)))  

(defmethod other-turn (( self game))
  (setf (turn self)
	(if (eql (turn self) (black self))
	    (white self)
	    (black self))))

(defmethod get-event-from-queue (( self game))
  (qpop (event-queue self)))

(defmethod do-pending-events (( self game))
  (loop until (empty-p (event-queue self))
	do (do-one-event self)))

;; by default do nothing on an event
;; unless there is a specific handle-event method for the specific
;; event handler class.
(defmethod handle-event ( eh ( ev blocker-event))
  (declare (ignore eh))
  nil)

(defmethod dispatch-event (( self game) ( ev blocker-event))
  (dolist ( eh (cons self (event-handlers self)))
    (handle-event eh ev)))


(defmethod do-one-event (( self game))
  (let (( ev (get-event-from-queue self)))
    (if (legal-p ev)
	(dispatch-event self ev)
	(progn (format t "Invalid event: ~A~%" (printable ev))
	       (format t "~4@TValid events are:~%")
	       (dolist ( ev (events self))
		 (when (legal-p ev)
		   (format t "~4@T~4@T~A~%" (printable ev))))))))

(defmethod play (( self game))
  (format t "PLAYING: computer=~A~%" (computer self))
  (let (( num-moves 0))
    (loop
     do
     (incf num-moves)
     ;; (format t  "EVENT #~D/~D~%" num-moves (remaining-moves self))

     (cond
       ((> num-moves (remaining-moves self))
	(qflush (event-queue self))
	(draw self)
	(return-from play nil))

       ((game-over self)
	(return-from play nil))

       ((null (empty-p (event-queue self)))
	(do-one-event self))

       ((winner-p (turn self))
	;; in case we found a winning move before the queue was finished
	;; we need to flush the rest of the queue
	(qflush (event-queue self))
	(draw self)
	(return-from play (turn self)))

       ((looser-p (turn self))
	;; in case we found a winning move before the queue was finished
	;; we need to flush the rest of the queue
	(qflush (event-queue self))
	(draw self)
	(return-from play (opponent (turn self))))
	     
       ((member (turn self) (computer self))
	(draw self)
	(put-event self "ma"))
       
       (t
	(draw self)
	(select-event self))))))

(defmethod winner-of-game (( self game) &key verbose)
  (dolist ( player (list (opponent (turn self))
		       (turn self)))
    (when (winner-p player :verbose verbose)
      (return-from winner-of-game player)))
  nil)

(defmethod print-check-mate (( self game) ( player-in-check player))
  (format t "!!!!!!!!!!!!!!!!!!!!!!!!~%")
  (format t "!!!!!! CHECK MATE !!!!!!~%")
  (format t "!!!!!! ~A WINS !!!!!!~%"  (name (opponent player-in-check)))
  (format t "!!!!!!!!!!!!!!!!!!!!!!!!~%")
  (opponent player-in-check))

(defmethod find-direction (( self game) ( symbol string))
  (find-if #'(lambda ( dir) (equal symbol (-symbol dir)))
	   (directions self)))


(defmethod find-event-by-shortcut (( game game) ( short string))
  (or (find-if #'(lambda ( ev)
		   ;; must use equal with strings, not eql
		   (equal short (shortcut ev)))
	       (events game))
      (format t "no such event ~A~%" short)))

(defmethod save-history (( self game))
  nil)

(defmethod vec-direction (( self game) dx dy)
  (let* (( vec-angle (atan dy dx))
	 ( match-dir (car (directions self)))
	 ( match-angle (angle match-dir)))
    (dolist ( dir (cdr (directions self)))
      (when (< (abs (- (angle dir) vec-angle))
	       (abs (- match-angle vec-angle)))
	(setf match-dir dir
	      match-angle  (angle dir))))

    match-dir))

(defmethod swap-strategies (( self game))
  (loop for wsc in (components (strategy (white self)))
	for bsc in (components (strategy (black self)))
	do
	(rotatef (coef wsc) (coef bsc))))



(defmethod print-object (( self player) stream)
  (print-unreadable-object (self stream)
    (format stream "+~A" (name self))))

(defmethod printable (( self player))
  (name self))

(defmethod get-state-key (( self player))
  (cons self (get-state-key (blocker self))))

(defclass black ( player )
  (( name
     :initform "BLACK")
   ( symbol
     :initform '( "b" "B"))
   ))

(defmethod initialize-instance :after (( self black) &rest ignored)
  (declare (ignore ignored))
  (setf (home-row self) (1- (num-rows (blocker self)))))

(defmethod initial-placement (( self black))
  (let (( y (1- (num-rows (blocker self)))))

    (place-piece (godfather self) :x 5 :y y       :dx 1 :dy 1)
    (place-piece (frecher   self) :x 6 :y y       :dx 1 :dy 1)
    (place-piece (thin      self) :x 7 :y (- y 1) :dx 1 :dy 2)
    (place-piece (flat      self) :x 4 :y (- y 1) :dx 1 :dy 2)
    (place-piece (fat       self) :x 5 :y (- y 2) :dx 2 :dy 2)))
   
(defclass white ( player )
  (( name :initform "WHITE")
   ( symbol :initform '("w" "W"))))

(defmethod initialize-instance :after (( self white) &rest ignored)
  (declare (ignore ignored))
  (setf (home-row self) 0))

(defmethod initial-placement (( self white))
  (place-piece (godfather self) :x 5 :y 0 :dx 1 :dy 1)
  (place-piece (frecher   self) :x 4 :y 0 :dx 1 :dy 1)
  (place-piece (thin      self) :x 3 :y 0 :dx 1 :dy 2)
  (place-piece (flat      self) :x 6 :y 0 :dx 1 :dy 2)
  (place-piece (fat       self) :x 4 :y 1 :dx 2 :dy 2))

(defmethod winner-p (( self player) &key verbose)
  (assert (eql self (turn (blocker self))))
  (when (eql (y-coord (state (godfather self)))
	     (home-row (opponent self)))
    (return-from winner-p t))

  (when (and verbose
	     (is-check (opponent self)))
    (format t "~A: YOU ARE IN CHECK~%" (name (opponent self))))

  (null (any-valid-move (opponent self))))


(defmethod looser-p (( self player))
  (assert (eql self (turn (blocker self))))
  (unless (any-valid-move self)
    (format t "~A: YOU HAVE NO VALID MOVE~%" (name self))
    t ))

(defmethod is-valid-comb (( self player) ( piece piece) ( comb combination)
			  &key verbose)

  ;; TODO the return value of apply-combination could immediately
  ;;     be examined to see if it leaves in check without having to apply
  ;;     it again in the function leaves-in-check
  (when (apply-combination (state piece)
			   (blocker self)
			   (opponent self)
			   comb
			   nil
			   :verbose verbose)
    (null (leaves-in-check piece comb))))

(def-cached-method any-valid-move (( self player))
  (get-state-key self)

  (let ( save-turn)
    (setf save-turn (turn (blocker self))
	  (turn (blocker self)) self)
    (prog1 (any-valid-comb self)
      (setf (turn (blocker self)) save-turn))))

(def-cached-method any-valid-comb (( self player))
  (get-state-key self)
  
  (let ( comb)
    (dolist ( piece (pieces self))
      (when (setf comb (any-valid-comb piece))
	(return-from any-valid-comb
	  (make-instance 'piece-move :piece piece :combination comb))))
    nil))

(defmethod does-comb-attack (( self player)
			     ( defender piece)
			     ( comb combination))
  (find-if #'(lambda ( attacker)
	       (does-comb-attack attacker defender comb))
	   (pieces self)))

(defmethod is-check-mate (( self player) in-check-p)
  (and in-check-p
       (null (any-valid-move self))))

(defmethod move-automatic (( self player))
  (let (( pm (calc-best-move self)))
    (format t "~A move ~D/~A: ~A" (name self)
	    (volume (piece pm))
	    (name (piece pm))
	    (printable (combination pm)))
    (move-piece (piece pm) (combination pm))))

(defmethod move-interactive (( self player))
  (let (( piece (choose-movable self))
	history
	dir
	)
    (when (equal "undo" piece)
      (other-turn (blocker self))
      (return-from move-interactive "undo"))

    (unless piece
      (format nil "~A: no piece is movable~%" (name self))
      (assert piece))

    (format t "~A ~D~%" (printable piece) (number-of-rolls piece))

    (dotimes ( i (number-of-rolls piece))
      (when (plusp i)
	(draw (blocker piece)))
      (setf dir (choose-direction piece history i))
      (push dir history)
      (move-piece dir piece))
    
    (when (is-check self)
      (format t " ---------- INVALID MOVE -----------~%")
      (dolist ( dir history)
	(unmove dir piece))
      (draw (blocker piece))
      (return-from move-interactive "must-move"))
    
    nil))

    
(defmethod move-player (( self player))
  (do (( must-move t))
      ( must-move )
    (setf must-move nil)
    (if (member self (computer (blocker self)))
	(move-automatic self)
	(case (move-interactive self)
	  ( "undo"
	   (return-from move-player nil))

	  ( nil
	   nil)

	  ( "must-move"
	   (setf must-move t)))))

  (when (winner-p self :verbose t)
    (format t "WINNER = ~A~%" (printable self))
    self))
    
(defmethod print-rated (( pm piece-move) ( rating number))
  (format t "[~A ~4,2F]" (printable pm) rating))

(defmethod calc-best-move (( self player))

  (when (verbose (blocker self))
    (print "Thinking ..."))
  (let ( rating
	max-rating
	max-pm
	pm
	comb
	)
    (assert (pieces self))
    
    (dolist (piece (pieces self))

      (when (verbose (blocker self))
	(format t "~A" (printable piece))
	(force-output))

      (assert (plusp (hash-table-count (eff-comb-table piece))))
      
      (loop for key being the hash-keys of(eff-comb-table piece)
	    do
	    (when (verbose (blocker self))
	      (format t "." )
	      (force-output))

	    (when (setf comb (choose-comb-from-effective piece key))

	      (setf pm (make-instance 'piece-move :piece piece :combination comb))
	      (reduce-piece-move pm)
	      (move-piece piece comb)

	      (when (winner-p self)
		(unmove piece comb)
		(print "found winning move")
		(return-from calc-best-move pm))
	      
	      (setf rating (calc-position-rating self))

	      (unless (and max-rating
			   max-pm)
		(setf max-pm pm
		      max-rating rating))

	      (when (and rating
			 (> rating max-rating))
		(setf max-rating rating
		      max-pm pm)
		(when (verbose (blocker self))
		  (print-rated max-pm rating)))
	      
	      (unmove piece comb))))

    (when (verbose (blocker self))
      (terpri))

    (format t "suggesting combination of rating = ~F ~A~%"
	    max-rating
	    max-pm)
    
    max-pm))

(defmethod find-piece (( self player ) ( volume number))
  (find-if #'(lambda ( piece)
	       (= volume (volume piece)))
	   (pieces self)))


(def-cached-method is-check (( self player))
  (get-state-key self)
  
  (let ( comb)
    (dolist ( piece (pieces (opponent self)))
      (when (setf comb (is-attacking piece (godfather self)))
	(return-from is-check (make-instance 'piece-move
					     :piece piece
					     :combination comb))))
    nil))

(defmethod select-piece (( player player) ( piece piece))
  (setf (selected-piece player) piece))

(defmethod deselect-piece (( player player))
  (setf (selected-piece player) nil))

(defmethod init-selected-combination (( player player))
  (setf (selected-combination-conc player) (list nil)))

(defmethod remember-direction (( player player) ( dir direction))
  (tconc (selected-combination-conc player) dir))

(defmethod selected-combination (( player player))
  (car (selected-combination-conc player)))


(defmethod calc-position-rating (( self player))
  (calc-position-rating (strategy self)))


(def-cached-method get-shadow-and-wasted-shadow (( self player))
  (get-state-key self)
  (let (( shadow-hash (make-hash-table))
	( wasted-shadow 0))
    
    (dolist ( piece (pieces self))
      (loop for square being the hash-keys of (get-shadow piece)
	    do
	    (when (gethash square shadow-hash)
	      (incf wasted-shadow))
	    (setf (gethash square shadow-hash) t)))

    (list (hash-table-count shadow-hash) wasted-shadow)))

(defmethod wasted-shadow (( player player))
  (cadr (get-shadow-and-wasted-shadow player)))

(defmethod shadow-size (( player player))
  (car (get-shadow-and-wasted-shadow player)))

(defmethod computer-becomes (( self player))
  (let (( blo (blocker self)))
    (unless (member self (computer blo))
      (push self (computer blo)))))




(defmethod diagonal-p (( self direction))
  nil)

(defmethod ident-p (( self direction))
  nil)

(defmethod printable (( self direction))
  (format nil "+~A" (-symbol self)))

(defmethod print-object (( self direction) stream)
  (print-unreadable-object (self stream)
    (format stream "~A" (printable self))))

(defmethod roll (( self direction ) ( state state))
  nil)


(defmethod diagonal-p (( self diagonal))
  t)


(defmethod initialize-instance :after (( self diagonal) &rest ignored)
  (declare (ignore ignored))
  (loop for comp in (components self)
	do
	(incf (deltax self) (deltax comp))
	(incf (deltay self) (deltay comp))))

(defmethod ident-p (( self no-direction))
  t)

(defmethod inverse-p (( dir1 direction)
		      ( dir2 direction))
  (= 0
     (+ (deltax dir1) (deltax dir2))
     (+ (deltay dir1) (deltay dir2))))


(defmethod move-piece (( dir direction) ( piece piece))
  (move-generic dir piece "move"))

(defmethod unmove (( dir direction) ( piece piece))
  (move-generic dir piece "unmove"))

(defmethod move-generic (( dir direction) ( piece piece) ( parity string))
  (pick-up piece)
  (roll (if (equal "move" parity)
	     dir
	     (inverse dir))
	(state piece))
  (update (blocker piece) piece)
  (if (equal "move" parity)
      (remember-move (blocker piece) dir piece)
      (forget-move (blocker piece)))
  (save-history (blocker piece))
  (dolist ( player (list (opponent (player piece))
		       (player piece)))
    (when (is-check player)
      (return-from move-generic player)))

  ;; nil means this was not a winning move
  nil)

;;;    WEST    ;;;

(defmethod clockwise (( self west))
  (northwest))

(defmethod counterclockwise (( self west))
  (southwest))

(defmethod inverse (( self west))
  (east))

(defmethod roll (( self west) ( state state))
  (cond ((= 0 (dz state))
	 (decf (x-coord state)))

	(t
	 (decf (x-coord state) (dz state))
	 (rotatef (dz state) (dx state)))))

;;;    EAST    ;;;
(defmethod clockwise (( self east))
  (southeast))

(defmethod counterclockwise (( self east))
  (northeast))

(defmethod inverse (( self east))
  (west))

(defmethod roll (( self east) ( state state))
  (incf (x-coord state) (dx state))
  (unless (= 0 (dz state))
    (rotatef (dz state) (dx state))))
  

;;;    NORTH   ;;;

(defmethod clockwise (( self north))
  (northeast))

(defmethod counterclockwise (( self north))
  (northwest))

(defmethod inverse (( self north))
  (south))

(defmethod roll (( self north) ( state state))
  (incf (y-coord state) (dy state))
  (unless (= 0 (dz state))
    (rotatef (dz state) (dy state))))

;;;  SOUTH  ;;;

(defmethod clockwise (( self south))
  (southwest))

(defmethod counterclockwise (( self south))
  (southeast))

(defmethod inverse (( self south))
  (north))

(defmethod roll (( self south) ( state state))
  (cond
    ((= 0 (dz state))
     (decf (y-coord state)))

    (t
     (decf (y-coord state) (dz state))
     (rotatef (dz state) (dy state)))))

;;; NORTHEAST ;;;

(defmethod clockwise (( self northeast))
  (east))

(defmethod counterclockwise (( self northeast))
  (west))

(defmethod inverse (( self northeast))
  (southwest))

(defmethod roll (( self northeast) ( state state))
  (incf (y-coord state))
  (incf (x-coord state)))

;;; SOUTHEAST ;;;

(defmethod clockwise (( self southeast ))
  (south))

(defmethod counterclockwise (( self southeast))
  (east))

(defmethod inverse (( self southeast))
  (northwest))

(defmethod roll (( self southeast) ( state state))
  (decf (y-coord state))
  (incf (x-coord state)))


;;; NORTHWEST ;;;

(defmethod clockwise (( self northwest))
  (north))

(defmethod counterclockwise (( self northwest))
  (west))

(defmethod inverse (( self northwest))
  (southeast))

(defmethod roll (( self northwest) ( state state))
  (incf (y-coord state))
  (decf (x-coord state)))


;;; SOUTHWEST ;;;

(defmethod clockwise (( self southwest))
  (west))

(defmethod counterclockwise (( self southwest))
  (south))

(defmethod inverse (( self southwest))
  (northeast))

(defmethod roll (( self southwest) ( state state))
  (decf (y-coord state))
  (decf (x-coord state)))

;;; IDENTITY ;;;

(defmethod inverse (( self no-direction))
  (no-direction))


(defmethod sweep-directions (( self direction))
  (unless (swept-directions self)
    (setf (swept-directions self)
	  (list (clockwise self)
		self
		(counterclockwise self))))
  (swept-directions self))



