;;$Id: piece.lisp,v 1.8 2005/05/23 21:30:06 jimka Exp $

#|

 This software is Copyright (c) 2005 Jim Newton <jimka@rdrop.com>

 Jim Newton grants you the rights to distribute
 and use this software as governed by the terms
 of the Lisp Lesser GNU Public License
 (http://opensource.franz.com/preamble.html),
 known as the LLGPL.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
|#

(in-package blocker)

(defmethod get-state-key (( self piece))
  (cons self (get-state-key (blocker self))))

(defmethod godfather-p (( ignored piece))
  (declare (ignore ignored))
  nil)

(defmethod leaves-in-check (( self piece) ( comb combination))
  (when (predict-state self comb)
    (move-piece self comb)
    (prog1 (is-check (player self))
      (unmove self comb))))

(def-cached-method any-valid-comb (( self piece))
  (get-state-key self)
  (loop for key being the hash-keys of (eff-comb-table self)
	do
	(dolist ( comb (gethash key (eff-comb-table self)))
	  (when (is-valid-comb (player self) self comb :verbose nil)
	    (return-from any-valid-comb comb)))))

(defmethod calc-combinations (( piece piece))
  (let ( l-comb)

    (setf l-comb
	  (loop for comb in (all-combinations (number-of-rolls piece) piece)
		unless (redundant-p comb)
		collect comb))
    
    (setf (combinations piece ) l-comb)
    (calc-effective-combinations piece)))

(defmethod calc-canonical-comb-key (( piece regular-piece) comb)
  (declare (ignore piece))
  (effective-combination comb))

(defmethod calc-canonical-comb-key (( piece piece) comb)
  (declare (ignore piece))
  comb)

;; For two of the pieces Godfather and Frecher there are several
;; combinations that leave the piece in exactly the same state.
;; However it is possible that some of the combinations will not be
;; valid at certain times in the game because of obstructions.
;; This function builds a hash table (eff-comb-table piece).
;; The value at each key is a list of combinations. For godfather
;; and Frecher the list has length > 1. for the other pieces the
;; lengths are exactly 1.  But in each case each list contains all
;; the combintations (instances of class combination) which will leave
;; the piece in the same state.  The point is that when we are looking
;; for the best move, we need only look at the first one from each of
;; these lists which IS a valid move.  This brings the search space
;; down to a smaller size.
;;
;; In additions to this the godfather contains some diagonal directions
;; in its combinations.  The function comb-expand-diag returns a list
;; of combinations completely manhattan but equivalent in final state.
;; Including these in this hash table here means the list is longer,
;; but it does not have to be expanded later in the function
;; apply-combination which makes that function faster.  In addition
;; because we only calculate the rating of one combination of each
;; list there are fewer calculations to be performed.
(defmethod calc-effective-combinations (( piece piece))
  (dolist ( comb (combinations piece))
    (dolist ( manhattan (if (comb-contains-diagonal comb)
			    (comb-expand-diag comb)
			    (list comb)))
      (cond
	((find manhattan (gethash (calc-canonical-comb-key piece manhattan)
				  (eff-comb-table piece))))
	(t
 	 (push manhattan (gethash (calc-canonical-comb-key piece manhattan)
				  (eff-comb-table piece)))))))
  
  (loop for key being the hash-keys of (eff-comb-table piece)
	do (setf (gethash key (eff-comb-table piece))
	      (sort (gethash key (eff-comb-table piece))
		    #'(lambda ( a b)
			(< (length (directions a))
			   (length (directions b)))))))
  )

(defmethod all-combinations (( number-of-rolls number) ( piece piece))
  (loop for li in (all-combinations number-of-rolls (directions piece))
	collect (make-instance 'combination :directions li)))

(defmethod does-comb-attack (( attacker piece) ( defender piece) ( comb combination))
  (let (( blo (blocker attacker))
	new-state)
    
    (when (setf new-state (apply-combination (state attacker)
					     blo
					     (player defender)
					     comb
					     nil
					     :verbose nil))

      (member defender (get-occupants new-state)))))

(defmethod choose-direction (( piece piece) ( history list) ( i number))
  ;; WARNING history is in opposite order
  nil
  )
(defmethod place-piece (( self piece) &key x y dx dy)
  (setf (x-coord (state self)) x
	(y-coord (state self)) y
	(dx (state self)) dx
	(dy (state self)) dy
	(dz (state self)) (/ (volume self) dx dy))
  (update (blocker self) self))


(defmethod move-piece (( piece piece) ( comb combination))
  (dolist ( dir (directions comb))
    (move-piece dir piece)))

(defmethod unmove (( piece piece) ( list-of-directions list))
  (when list-of-directions
    (unmove piece (cdr list-of-directions))
    (unmove (car list-of-directions) piece )))    

(defmethod unmove (( piece piece) ( comb combination))
  (unmove piece (directions comb)))

(defmethod predict-state (( self piece) ( comb combination))
  (apply-combination (state self) (blocker self) (opponent (player self)) comb nil :verbose nil))

(defmethod pick-up (( self piece))
  (dolist ( square (squares self))
    (setf (occupant square) nil))
  (setf (squares self) nil))



(defclass godfather ( regular-piece )
  (( volume
     :initform 0)
   ( name
     :initform "godfather")
   ( number-of-rolls
     :initform 2)))

(defmethod godfather-p (( ignored godfather))
  (declare (ignore ignored))
  t)

(defmethod initialize-instance :after (( self godfather) &rest ignored)
  (declare (ignore ignored))
  ;;(format t "initializing godfather~%")
  (setf (directions self) (directions (blocker self)))
  (calc-combinations self))

;;size = [1,1,1]
(defclass frecher ( rollable regular-piece)
  (( volume :initform 1)
   ( name :initform "frecher")
   ( number-of-rolls :initform 3)))

(defclass thin ( rollable )
  (( volume :initform 2)
   ( name :initform "thin")
   ( number-of-rolls :initform 2)))

(defclass flat ( rollable )
  (( volume          :initform 4)
   ( name            :initform "flat")
   ( number-of-rolls :initform 2)))

(defclass fat ( rollable regular-piece)
  (( volume :initform 8)
   ( name :initform "fat")
   ( number-of-rolls :initform 1)))

 
(defmethod is-attacking (( attacker piece) ( defender piece))
  (loop for comb in (combinations attacker)
	do
	(when (does-comb-attack attacker defender comb)
	  (return-from is-attacking comb)))
  nil)


(defmethod does-comb-attack (( attacker piece) ( defender piece) ( comb combination))
  (let (( new-state (apply-combination (state attacker)
				       (blocker attacker)
				       (player defender)
				       comb
				       nil
				       :verbose nil)))
    
    (and new-state
	 (member defender (get-occupants new-state)))))


;; are two lists equal up to the number of elements in the
;; first of the two lists.
;; E.g.,  (1 2 3 ) == ( 1 2 3 4 5)
(defmethod matches-first-n (( partial-list1 list) ( list2 list))
  (cond
    ;; they are equal if both lists are nil
    ((null (or partial-list1 list2))
     t)

    ;; if we have exausted the first list, then YES
    ((null partial-list1)
     t)

    ;; if list1 is longer than list2 then NO
    ((null list2)
     nil)

    ;; if the first elements are not equal then NO
    ((null (eql (car partial-list1) (car list2)))
     nil)

    ;; if the rest of the elements are not equal then NO
    (t
     (matches-first-n (cdr partial-list1) (cdr list2)))))

(defmethod is-legal-direction (( self piece)
			       ( direction direction)
			       ( partial-list-of-dirs list))
  (find-if #'(lambda ( comb)
	       (apply-combination
		(state self) (blocker self) (opponent (player self)) comb
		partial-list-of-dirs
		:verbose nil
		))
	   (find-matching-combinations self
				       `(,@partial-list-of-dirs ,direction))))

(defmethod find-matching-combinations (( self piece) ( partial-target-comb list))
  (setof comb (combinations self)
	 (matches-first-n partial-target-comb (directions comb))))

(defmethod printable (( piece piece))
  (format nil "~A~D" (car (-symbol (player piece))) (volume piece)))

(defmethod print-object (( piece piece) stream)
  (print-unreadable-object (piece stream)
    (format stream "+~A" (printable piece))))

(defmethod dist (( self piece) (piece piece))
  (dist (state self) (state piece)))

(def-cached-method get-shadow (( self piece))
  (get-state-key self)
  (let (( dict (make-hash-table))
	( blo (blocker self))
	state
	)
    (dolist ( square (squares self))
      (setf (gethash square dict) t))

    (dolist ( comb (combinations self))
      (when (setf state (apply-combination (state self)
					   blo
					   (opponent (player self))
					   comb
					   nil
					   :verbose nil))
	(dolist ( square (get-squares state))
	  (setf (gethash square dict) t))))

    dict))
    
(defmethod closeness-to-backline (( self piece))
  (abs (- (home-row (player self))
	  (y-coord (state self)))))

(defmethod approach-direction (( self piece) ( piece piece))
  (let (( v1 (center self))
	( v2 (center piece)))

    (vec-direction (blocker self)
		   (- (car  v2) (car  v1))
		   (- (cadr v2) (cadr v1)))))

(defmethod center (( self piece))
  (center (state self)))



(defmethod show-combinations (( blo blocker))
  (mapcar #'show-combinations (players blo)))

(defmethod show-combinations (( player player))
  (mapcar #'show-combinations (pieces player)))

(defmethod show-combinations (( piece piece))
  (show-basic-combinations piece)
  (show-eff-combinations piece)
  nil)

(defmethod show-basic-combinations (( piece piece))
  (dolist ( comb (combinations piece))
    (format t "~A: act ~A~%" (name piece) comb)))

(defmethod show-eff-combinations (( piece piece))
  (loop for key being the hash-keys of (eff-comb-table piece)
	do
	(format t "~A: eff ~A: ~A~%" (name piece) key (gethash key (eff-comb-table piece)))))


(def-cached-method choose-comb-from-effective (( piece piece) key)
  (cons key (get-state-key piece))
  (find-if #'(lambda (comb)
	       (is-valid-comb (player piece) piece comb :verbose nil))
	   (gethash key (eff-comb-table piece))))
  


(defmethod len (( self combination))
  (length (directions self)))

(defmethod all-combinations (( x-length (eql 1)) ( list-of-directions list))
  (mapcar #'list list-of-directions))
  
(defmethod all-combinations (( x-length integer) ( list-of-directions list))
  (loop for li in (all-combinations (1- x-length) list-of-directions)
	nconcing
	(loop for dir in list-of-directions
	      collect (cons dir li))))

(defmethod comb-to-list (( self combination))
  (loop for dir in (directions self)
	collect (-symbol dir)))

(defmethod printable (( self combination))
  (format nil "~A" (comb-to-list self)))

(defmethod print-object (( self combination) stream)
  (print-unreadable-object (self stream)
    (format stream "~A" (printable self))))

(defmethod redundant-p (( comb combination))
  (assert (directions comb))
  (assert (car (directions comb)))
  (cond

    ((ident-p (car (directions comb)))
     t)

    ((and (= 3 (length (directions comb)))
	  (ident-p (nth 1 (directions comb)))
	  (null (ident-p (nth 2 (directions comb)))))
     t)

    ((exists dir (directions comb)
	     (and (null (ident-p dir))
		  (exists dir2 (directions comb)
			  (inverse-p dir dir2))))
     t)

    (t
     nil)))

(defmethod effective-combination (( comb1 combination))
  (let (( deltax 0)
	( deltay 0))
    (loop for dir in (directions comb1)
	  do
	  (incf deltax (deltax dir))
	  (incf deltay (deltay dir)))

    (list deltax deltay)))

(defmethod equiv-combinations (( comb1 combination) ( comb2 combination))
  (eql (effective-combination comb1)
       (effective-combination comb2)))

(defmethod comb-contains-diagonal (( combination combination))
  (find-if #'diagonal-p (directions combination)))

(defmethod comb-expand-diag (( comb combination))
  (loop for l in (comb-expand-diag (directions comb))
	collect (make-instance 'combination
			       ;; :piece (piece comb)
			       :directions l)))

(defmethod comb-expand-diag (( list-of-dirs (eql nil)))
  nil)

(defun all-permutations ( items)
  (cond
    ((null items) nil)
    ((null (cdr items)) (list items))
    (t
     (loop for item in items
	   nconc
	   (loop for remaining in (all-permutations (remove item items :count 1))
		 collect (cons item remaining))))))

(defmethod comb-expand-diag (( dir diagonal))
  (all-permutations (components dir)))

(defmethod comb-expand-diag (( dir direction))
  (list (list dir)))

(defmethod comb-expand-diag (( self (eql nil)))
  nil)

(defmethod comb-expand-diag (( list-of-dirs list))
  (let (( begin-lists (comb-expand-diag (car list-of-dirs)))
	( end-lists   (comb-expand-diag (cdr list-of-dirs)))
	)
    (cond
      ((null end-lists)
       begin-lists)
		    
      (t
       (loop for begin-list in begin-lists
	     nconc (loop for end-list in end-lists
			 collect (append begin-list end-list)))))))



(defmethod printable (( self piece-move))
  (format nil "~A`~A" (printable (piece self)) (printable (combination self))))

(defmethod print-object (( self piece-move) stream)
  (print-unreadable-object (self stream)
    (format stream "~A ~A" (piece self) (combination self))))

(def-cached-method reduce-combination (( piece piece) (comb combination))
  (cons (directions comb) (get-state-key piece))
  
  (find-if #'(lambda ( co)
	       (and (equal (calc-canonical-comb-key piece co)
			   (calc-canonical-comb-key piece comb))
		    (is-valid-comb (player piece) piece co :verbose nil)))
	   (combinations piece)))
  
(defmethod reduce-piece-move (( self piece-move))
  (setf (combination self) (reduce-combination (piece self) (combination self)))
  (assert (combination self))
  self)
  

;; methods of state

(defmethod print-object (( self state) stream)
  (print-unreadable-object (self stream)
    (format stream "~A,~A:~Dx~Dx~D"
	    (x-coord self)
	    (y-coord self)
	    (dx self)
	    (dy self)
	    (dz self))))

(defmethod shorthand (( self state))
  ;;(cons (volume (piece self))
  (shorthand (piece self))
  ;;)
  )

(defmethod shorthand (( self regular-piece))
  (compress (list (x-coord (state self))
		  (y-coord (state self)))
	    (list (num-cols (blocker self))
		  (num-rows (blocker self)))))

(defmethod shorthand (( self piece))
  (compress (list (x-coord (state self))
		  (y-coord (state self))
		  (1- (dx (state self)))
		  (1- (dy (state self))))
	    (list (num-cols (blocker self))
		  (num-rows (blocker self))
		  1
		  1)))

(defun compress ( l-val l-max)
  (cond
    ((null l-val)
     0)
    (t
     (+ (car l-val)
	(* (1+ (car l-max)) (compress (cdr l-val) (cdr l-max)))))))

(defmethod apply-combination (( self state)
			      ( blo blocker)
			      ( opponent player)
			      ( combination combination)
			      ( partial-list-of-dirs list)
			      &key verbose)
  (when verbose
    (format t "considering ~A already having moved ~A~%" combination
	    partial-list-of-dirs))
  (apply-combination self blo opponent
		     (make-instance 'combination
				    :directions (nthcdr (length partial-list-of-dirs)
							(directions combination)))
		     nil
		     :verbose verbose))
  

(defmethod apply-combination (( self state)
			      ( blo blocker)
			      ( opponent player)
			      ( combination combination)
			      ( partial-list-of-dirs (eql nil))
			      &key verbose)
  
  (when verbose
    (format t "considering ~A~%" combination))
  
  (let ( new-state)
    (when (comb-contains-diagonal combination)
      (dolist ( comb (comb-expand-diag combination))

	;; RECURSIVE CALL
	(setf new-state (apply-combination self blo opponent comb
					   nil
					   :verbose verbose))
	(when new-state
	  (when verbose (format t "case 1~%"))
	  (return-from apply-combination new-state))
	)
      (when verbose  (format t "case 2~%"))
      (return-from apply-combination nil))

    (setf new-state (duplicate self))
    (dolist ( dir (directions combination))
      (roll dir new-state)
      (unless (valid-intermediate-state new-state opponent)
	(when verbose	(format t "case 3~%"))
	(return-from apply-combination nil)))
    (when verbose  (format t "case 4~%"))
    (return-from apply-combination new-state)))

(defmethod valid-intermediate-state (( new-state state) ( opponent player))
  (unless (get-squares new-state)
    (return-from valid-intermediate-state nil))

  (let (( occupants (get-occupants new-state)))
    ;; if there are no occupants of one of the preposed squares, then
    ;;   the state is valid.
    (unless occupants
      (return-from valid-intermediate-state t))

    ;; if there are occupants, but the only occupant is the piece itself,
    ;;   then also consider it valid
    (setf occupants (delete (piece new-state) occupants))
    (unless occupants
      (return-from valid-intermediate-state t))

    ;; when there is a piece in one of the squares other than the opponent's
    ;; godfather, then it is NOT valid.
    (when (delete (godfather opponent) occupants)
      (return-from valid-intermediate-state nil))
    
    ;; is godfather attacking godfather?
    (when (eql (piece new-state)
	       (godfather (player (piece new-state))))
      (return-from valid-intermediate-state nil))

    t))

  
(defmethod duplicate (( self state))
  (make-instance 'state
		 :piece (piece self)
		 :x-coord (x-coord self)
		 :y-coord (y-coord self)
		 :dx (dx self)
		 :dy (dy self)
		 :dz (dz self)))

(defmethod get-squares (( self state))
  (let (( blo (blocker (piece self))))
    (loop for col from 0 below (dx self)
	  nconcing
	  (loop for row from 0 below (dy self)
		;; get-square returns the square obj if it is a valid square
		;;    and returns nil if it is not
		collect (or (get-square blo (+ row (y-coord self)) (+ col (x-coord self)))
			    (return-from get-squares nil))))))

(defmethod get-occupants (( self state))
  (loop for square in (get-squares self)
	nconcing (when (occupant square)
		   (list (occupant square)))))

(defmethod center (( self state))
  (list (+ (x-coord self)
	   (/ (dx self) 2.0))
	(+ (y-coord self)
	   (/ (dy self) 2.0))))

(defmethod dist (( self state) ( state state))
  (let (( pt1 (center self))
	( pt2 (center self)))

    (+ (abs (- (car pt1) (car pt2)))
       (abs (- (cadr pt1) (cadr pt2))))))
