;; $Id: polyhedra.lisp,v 1.13 2005/05/23 21:30:06 jimka Exp $

#|

 This software is Copyright (c) 2005 Jim Newton <jimka@rdrop.com>

 Jim Newton grants you the rights to distribute
 and use this software as governed by the terms
 of the Lisp Lesser GNU Public License
 (http://opensource.franz.com/preamble.html),
 known as the LLGPL.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
|#

(in-package blocker)

(defclass point-3d ( blocker-base )
  (( vec
     :reader vec
     :initform (make-array '(3)
			   :element-type 'number
			   :initial-element 0.0
			   :adjustable nil))))
(defun sqr ( x )
  (* x x))

(defmethod dist (( pt1 point-3d) ( pt2 point-3d))
  (sqrt (+ (sqr (- (x-coord pt1) (x-coord pt2)))
	   (sqr (- (y-coord pt1) (y-coord pt2)))
	   (sqr (- (z-coord pt1) (z-coord pt2))))))

(defmethod mid-point (( pt1 point-3d) ( pt2 point-3d))
  (scale-point (add-point pt1 pt2) 0.5))

(defmethod scale-point (( scale number) ( pt point-3d))
  (scale-point pt scale))

(defun make-3d-point ( x y z)
  (let (( pt (make-instance 'point-3d)))
    (setf (x-coord pt) x
	  (y-coord pt) y
	  (z-coord pt) z)
    pt))


(defmethod scale-point (( pt point-3d) ( scale number))
  (make-3d-point (* scale (x-coord pt))
		 (* scale (y-coord pt))
		 (* scale (z-coord pt))))

(defmethod add-point (( pt1 point-3d) ( pt2 point-3d))
  (make-3d-point (+ (x-coord pt1) (x-coord pt2))
		 (+ (y-coord pt1) (y-coord pt2))
		 (+ (z-coord pt1) (z-coord pt2))))

(defmethod print-object (( self point-3d) stream)
  (print-unreadable-object (self stream :type t)
    (format stream "~A ~A ~A"
	    (x-coord self)
	    (y-coord self)
	    (z-coord self))))

(defmethod x-coord (( self point-3d))
  (aref (vec self) 0))

(defmethod y-coord (( self point-3d))
  (aref (vec self) 1))

(defmethod z-coord (( self point-3d))
  (aref (vec self) 2))

(defmethod (setf x-coord) (new-value (obj point-3d))
  (setf (aref (vec obj) 0) new-value)) 

(defmethod (setf y-coord) (new-value (obj point-3d))
  (setf (aref (vec obj) 1) new-value)) 

(defmethod (setf z-coord) (new-value (obj point-3d))
  (setf (aref (vec obj) 2) new-value)) 



(defclass matrix ( blocker-base)
  (( mat
     :reader mat
     :initform (make-array '( 3 3)
			   :initial-element 0.0
			   :initial-element 0.0
			   :adjustable nil))))

(defmethod print-object (( obj matrix) stream)
  (print-unreadable-object (obj stream :type t)
    (format stream "~A" (mat obj))))

(defmethod apply-transform (( pt point-3d) ( matrix matrix))
  (let* (( answer (make-instance 'point-3d))
	 ( new (vec answer)))
    (loop for row from 0 below 3
	  do
	  (loop for col from 0 below 3
		do
		(incf (aref new row)
		      (* (aref (vec pt) col)
			 (aref (mat matrix) row col)))))
					    
    answer))

;; destructively translate a point by a dx,dy
(defmethod apply-translate (( self point-3d) &key ( dx 0) ( dy 0))
  (incf (x-coord self) dx)
  (incf (y-coord self) dy)
  self)

(defmethod apply-rotate (( pt point-3d) ( mat matrix))
  (let (( answer (apply-transform pt mat)))
    (setf (x-coord pt) (x-coord  answer)
	  (y-coord pt) (y-coord  answer)
	  (z-coord pt) (z-coord  answer)))
  pt)
    

;; destructively round x, y, and z coords
(defmethod apply-round (( self point-3d))
  (loop for row from 0 below 3
	do
	(setf (aref (vec self) row) (round (aref (vec self) row))))
  self)

(defmethod apply-round (( self matrix))
  (loop for row from 0 below 3
	do
	(loop for col from 0 below 3
	      do
	      (setf (aref (mat self) row col)
		    (round (aref (mat self) row col)))))
  self)

(defmethod set-matrix (( mat matrix) value &key row col)
  (setf (aref (mat mat) row col) value))

(defun rotate-about-x ( theta)
  (let (( new (make-instance 'matrix))
	( c (cos theta))
	( s (sin theta)))
    
    (set-matrix new 1  :row 0 :col 0)
    (set-matrix new c  :row 1 :col 1)
    (set-matrix new (- s) :row 1 :col 2)
    (set-matrix new s  :row 2 :col 1)
    (set-matrix new c  :row 2 :col 2)
    new))

(defun rotate-about-y ( theta)
  (let (( new (make-instance 'matrix))
	( c (cos theta))
	( s (sin theta)))
    
    (set-matrix new c  :row 0 :col 0)
    (set-matrix new s  :row 0 :col 2)
    (set-matrix new 1  :row 1 :col 1)
    (set-matrix new (- s) :row 2 :col 0)
    (set-matrix new c  :row 2 :col 2)
    new))

(defun rotate-about-z ( theta)
  (let (( new (make-instance 'matrix))
	( c (cos theta))
	( s (sin theta)))
    
    (set-matrix new  c :row 0 :col 0)
    (set-matrix new (- s) :row 0 :col 1)
    (set-matrix new  s :row 1 :col 0)
    (set-matrix new  c :row 1 :col 1)
    (set-matrix new  1 :row 2 :col 2)
    new))

(defmethod matrix-multiply-2 (( self matrix) (other matrix))
  (let (( new (make-instance 'matrix)))
    
    (loop for i from 0 below 3
	  do
	  (loop for j from 0 below 3
		do
		(set-matrix new 
			    (loop for k from 0 below 3
				  summing  (* (aref (mat self) j k)
					      (aref (mat other) k i)))
			    :row j
			    :col i)))
    new))

(defmethod matrix-multiply (( mat1 matrix) (mat2 matrix) &rest matrices)
  (cond
    ((null matrices)
     (matrix-multiply-2 mat1 mat2))

    (t
     (matrix-multiply-2 mat1
			(apply #'matrix-multiply mat2 matrices)))))

(defun rotate-3-angle ( alpha beta theta)
  (matrix-multiply (rotate-about-y (- beta))
		   (rotate-about-z (- alpha))
		   (rotate-about-y theta)
		   (rotate-about-z alpha)
		   (rotate-about-y beta)))

(defun unrotate-3-angle ( alpha beta theta)
  (rotate-3-angle alpha beta (- theta)))

(defun roll-N-mat ()
  (apply-round (rotate-about-x (* pi -90/180))))
(defun roll-S-mat ()
  (apply-round (rotate-about-x (* pi  90/180))))
(defun roll-E-mat ()
  (apply-round (rotate-about-y (* pi  90/180))))
(defun roll-W-mat ()
  (apply-round (rotate-about-y (* pi -90/180))))

(let (hash)
  (defun roll-mat (dir)
    (unless hash
      (setf hash (make-hash-table :size 4)
	    (gethash (north) hash) (roll-N-mat)
	    (gethash (south) hash) (roll-S-mat)
	    (gethash (west)  hash) (roll-W-mat)
	    (gethash (east)  hash) (roll-E-mat)))
    (gethash dir hash)))
      


;; This class represents a collection of squares.
;; The squares are considered collectively as a unit.  If they
;; get rendered on some viewer such as LTK they should all
;; be the same color.
(defclass face-list ( blocker-base )
  (( faces
     :initarg :faces
     :type array
     :accessor faces)))

;; This class models a 6 sided rectangular solid.
;; or perhaps (in the case of the godfather piece, a 1 sided piece.
;; The faces slot is a list of objects of class FACE.
;; and points is a uniq list of points (eq to those in the faces).
(defclass polyhedran ( face-list )
  (
   ;; uniq list of all the points in the faces
   ;; we will destructively modify the car and cadr of these points
   ;; to effectively move the pieces around the board
   ( points
     :initform nil
     :initarg :points
     :accessor points)))

(defmethod initialize-instance :after (( self polyhedran) &rest ignored)
  (declare (ignore ignored))
  (loop for face across (faces self)
	do
	(loop for pt across (points face)
	  do
	  (unless (member pt (points self))
	    (push pt (points self)))))
  
  (cond
    ((eql 1 (length (faces self)))
     (correlate-faces (aref (faces self) 0)
		      (aref (faces self) 0)))
    
    ((eql 0 (length (faces self)))
     (error "unexpected line 40 reached"))

    ;; set 0'th and 1st face as corresponding
    ;;     2'nd and 3rd
    ;;     4'th and 5th etc...
    ((evenp (length (faces self)))
     (loop for i from 0 below (/ (length (faces self)) 2)
	   do (correlate-faces (aref (faces self) (* 2 i))
			       (aref (faces self) (1+ (* 2 i))))))
    
    ( t (error "unexpected line 49 reached"))))

(defclass face ( blocker-base )
  (( points
     :type array
     :initarg :points
     :accessor points)
   ( color
     :initarg :color
     :initform (error "no color specified")
     :accessor color)))

(defclass polyhedran-face ( face )
  (( opposite-face
     :initform nil
     :type polyhedran-face
     :accessor opposite-face)))

(defmethod correlate-faces (( face1 polyhedran-face)
			    ( face2 polyhedran-face))
  (setf (opposite-face face1) face2
	(opposite-face face2) face1))

(defmethod correlate-faces (( face1 face)
			    ( face2 face))
  nil)

(defmethod center (( face face))
  (mid-point (aref (points face) 0)
	     (aref (points face) 2)))

(defun make-face ( color fc p1 p2 p3 p4)
  (make-instance fc
		 :color color
		 :points (make-array '( 4)
				     :element-type 'point-3d
				     :adjustable nil
				     :initial-contents (list p1 p2 p3 p4))))

(defmethod face-class (( fc (eql 'polyhedran)))
  'polyhedran-face)

(defmethod face-class (( fc (eql 'face-list)))
  'face)

;; class is either 'polyhedran or 'face-list
(defmethod make-faces (( class symbol) color point-lists)
  (make-instance class
		 :faces
		 (make-array (list (length point-lists))
			     :element-type (face-class class)
			     :adjustable nil
			     :initial-contents
			     (loop for pl in point-lists
				   collect (apply #'make-face (cons color
								    (cons (face-class class) pl)))))))

(defmethod make-specific-polyhedran (color
				     ( dz (eql 0))
				     ( p1 point-3d)
				     ( p2 point-3d)
				     ( p3 point-3d)
				     ( p4 point-3d))
  (make-faces 'polyhedran
	      color
	      (list (list p1 p2 p3 p4))))

(defmethod make-specific-polyhedran (color
				     ( dz number)
				     ( p1 point-3d)
				     ( p2 point-3d)
				     ( p3 point-3d)
				     ( p4 point-3d))
  (let* (( k-hat (make-3d-point 0 0 dz))
	 ( p5 (add-point p1 k-hat))
	 ( p6 (add-point p2 k-hat))
	 ( p7 (add-point p3 k-hat))
	 ( p8 (add-point p4 k-hat)))

    (make-faces 'polyhedran
		color
 		(list
		 ;; bottom
		 (list p1 p2 p3 p4)
		 
		 ;; top
		 (list p5 p6 p7 p8)
				
		 ;; left
		 (list p1 p4 p8 p5)
				
		 ;; right
		 (list p2 p3 p7 p6)
				
		 ;; front
		 (list p1 p2 p6 p5)
				
		 ;; posterior
		 (list p4 p3 p7 p8)))))

(defmethod make-specific-polyhedran (color ( x number) ( y number) dx dy dz)
  (let* (( p1 (make-3d-point x y 0))
	 ( i-hat (make-3d-point dx 0 0))
	 ( j-hat (make-3d-point 0 dy 0))
	 ( p2 (add-point p1 i-hat))
	 ( p3 (add-point p2 j-hat))
	 ( p4 (add-point p1 j-hat)))
    
    (make-specific-polyhedran color dz p1 p2 p3 p4)))

(defmethod make-polyhedran (( piece piece) color)
  (make-polyhedran (state piece) color))

(defmethod make-polyhedran (( state state) color)
  (let* (( p1 (make-3d-point (x-coord state) (y-coord state) 0))
	 ( i-hat (make-3d-point (dx state) 0 0))
	 ( j-hat (make-3d-point 0 (dy state) 0))
	 ( p2 (add-point p1 i-hat))
	 ( p3 (add-point p2 j-hat))
	 ( p4 (add-point p1 j-hat)))
    
    (make-specific-polyhedran color (dz state) p1 p2 p3 p4)))

(defmethod poly-roll (( phd polyhedran) x-bound y-bound ( mat matrix))
  (apply-translate phd :dx (- x-bound) :dy (- y-bound))
  (apply-rotate phd mat)
  (apply-translate phd :dx x-bound :dy y-bound))

(defmethod calc-extreme (( phd polyhedran) coord-fun extreme-fun)
  (let (( extreme (funcall coord-fun (car (points phd)))))
    (dolist ( pt (cdr (points phd)))
      (setf extreme (funcall extreme-fun extreme
			     (funcall coord-fun pt))))
    extreme))

(defmethod calc-dir-bound (( dir north) ( phd polyhedran))
  (calc-extreme phd #'y-coord #'max))

(defmethod calc-dir-bound (( dir south) ( phd polyhedran))
  (calc-extreme phd #'y-coord #'min))

(defmethod calc-dir-bound (( dir east) ( phd polyhedran))
  (calc-extreme phd #'x-coord #'max))

(defmethod calc-dir-bound (( dir west) ( phd polyhedran))
  (calc-extreme phd #'x-coord #'min))

(defmethod roll (( dir no-direction) ( phd polyhedran))
  phd)

(defmethod roll (( dir direction) ( phd polyhedran))
  (cond
    ((eql 1 (length (faces phd)))
     (apply-translate phd :dx (deltax dir) :dy (deltay dir)))
    
    (t
     (poly-roll phd
		(abs (* (deltax dir)
			(calc-dir-bound dir phd)))
		(abs (* (deltay dir)
			(calc-dir-bound dir phd)))
		(roll-mat dir)))))

(defmethod apply-translate (( phd polyhedran) &key (dx 0) (dy 0))
  (dolist ( pt (points phd))
    (apply-translate pt :dx dx :dy dy))
  phd)

(defmethod apply-rotate (( phd polyhedran) ( mat matrix))
  (dolist ( pt (points phd))
    (apply-rotate pt mat))
  phd)


(defclass virtual-3d ( blocker-base)
  nil)

(defclass virtual-piece ( virtual-3d)
  (( piece
     :initarg :piece
     :accessor piece)
   ( polyhedran
     :initarg :polyhedran
     :type polyhedran
     :accessor polyhedran)))

(defclass virtual-3d-equipment ( virtual-3d )
  (( blocker
     :accessor blocker
     :initarg :blocker)
   ( board-background
     :initarg :board-background
     :type face-list
     :accessor board-background)
   ( board-grid
     :initarg :board-grid
     :type face-list
     :accessor board-grid)
   ( virtual-pieces
     ;; type list
     :initarg :virtual-pieces
     :accessor virtual-pieces)))

(defmethod initialize-instance :after (( self virtual-3d-equipment)
				       &rest ignored)
  (declare (ignore ignored))
  (let (( color 0))
    
    (setf

     (board-background self)
     (make-faces 'face-list
		 color
		 (list (list (make-3d-point 0
					    0
					    0)
			     (make-3d-point (num-rows (blocker self))
					    0
					    0)
			     (make-3d-point (num-rows (blocker self))
					    (num-cols (blocker self))
					    0)
			     (make-3d-point 0
					    (num-cols (blocker self))
					    0)))))

    (incf color)

    (setf
     (board-grid self)
     (make-faces 'face-list
		 color
		 (loop for i from 0 by 2 to (* (num-rows (blocker self))
					       (num-cols (blocker self)))
		       collect (list (make-3d-point
				      (mod i (num-cols (blocker self)))
				      (truncate i (num-cols (blocker self)))
				      0)
				     (make-3d-point
				      (1+ (mod i (num-cols (blocker self))))
				      (truncate i (num-cols (blocker self)))
				      0)
				     (make-3d-point
				      (1+ (mod i (num-cols (blocker self))))
				      (1+ (truncate i (num-cols (blocker self))))
				      0)
				     (make-3d-point
				      (mod i (num-cols (blocker self)))
				      (1+ (truncate i (num-cols (blocker self))))
				      0)))))
     
    (incf color)
    
    (setf
     (virtual-pieces self)
     (loop for player in (players (blocker self))
	   for color2 from color to (1+ color)
	   nconcing
	   (loop for piece in (pieces player)
		 collect (make-instance 'virtual-piece
					:piece piece
					:polyhedran
					(make-polyhedran piece color2)))))))

(defmethod handle-event (( eh virtual-3d-equipment) ( ev blocker-event))
  nil
  )

(let ( phd )
  (defmethod handle-event (( eh virtual-3d-equipment) ( ev move-direction ))
    (let (( dir (find-direction (blocker eh) (shortcut ev))))
      (roll dir phd)))
    
  (defmethod handle-event (( eh virtual-3d-equipment) ( ev move-piece ))
    (let (( piece (find-piece (turn (blocker eh)) (volume ev)))
	  virtual-piece)
      ;; find the polyhedran that corresponds to the virtual-piece
      ;; pointing to this piece
      (setf virtual-piece (find-if #'(lambda ( vp)
				       (eq piece (piece vp)))
				   (virtual-pieces eh))
	    phd (polyhedran virtual-piece)))))
